package View;

use Curses;

# Set up interrupt handling
        #$#sigstack = 100; #preallocate some space
        #$sigkey{"INT"} = 3; #^C
        #$sigkeyf{"INT"} = 3;
        #$SIG{"INT"} = \&catch_int;
        #$sigkey{"KILL"} = 21; #^U
        #$sigkeyf{"KILL"} = 21;
        #$SIG{"KILL"} = \&catch_int;
        #$sigkey{"STOP"} = 21; #^S
        #$sigkeyf{"STOP"} = 21;
        #$SIG{"STOP"} = \&catch_int;
        $SIG{"WINCH"} = \&catch_int;

# Setup the terminal
        $win = new Curses; #::initsrc;
        #Curses::cbreak;    #interactive mode
        Curses::raw;    #interactive mode
        Curses::noecho;    #
        Curses::nodelay(1);    #
        Curses::nonl;      #trap return key
        Curses::intrflush(Curses::stdscr, 0);
        #Curses::meta(1); #8 bits
        Curses::keypad(1);
        #Curses::clear;

#        Curses::refresh;
#        Curses::clear;
Curses::getmaxyx($HEIGHT,$WIDTH);

END { Curses::endwin; }

####################################
# Class Variables

$FULL;     # The large window
$MINI;     # The tiny window
$EDITVIEW; # The window with the cursor

####################################
# Class Methods

sub new {
	my($class) = shift;
	my($style) = shift; # full or mini
	my($buf) = shift;
	#_DebugLog("View new $class $style $buf\n");
	my($self) = {};
	bless $self, $class;
	$self->{'buffer'} = $buf;
	#_DebugLog("View $self holds buffer $buf\n");
	$buf->RegisterAsObserver($self);

	if($style eq "mini"){
		$self->{"line"} = $HEIGHT - 1; # vertical location of view
		$self->{"height"} = 1;         # height of view
		$self->{"col"} = 0;            # horizontal location of view
		$self->{"width"} = $WIDTH;     # width of view
		$self->{"curcol"} = $buf->{"curcol"};
		$self->{"curline"} = $buf->{"curline"};
		$MINI->{'buffer'}->UnregisterAsObserver($MINI)
			if(defined $MINI);
		$MINI=$self;
	}
	if($style eq "full"){
		$self->{"line"} = 0;
                $self->{"height"} = $HEIGHT-1;
                $self->{"col"} = 0;
                $self->{"width"} = $WIDTH;
		$self->{"curcol"} = $buf->{"curcol"};
		$self->{"curline"} = $buf->{"curline"};
		$FULL->{'buffer'}->UnregisterAsObserver($FULL)
			if(defined $FULL);
		$FULL=$self;
	}

	return $self;
}

sub AssignToEditView {
	my($class)=shift;
	my($buf) = shift;
	#_DebugLog("View AssignToEditView $buf\n");
	$EDITVIEW->{'buffer'}->UnregisterAsObserver($EDITVIEW);
	$EDITVIEW->{'buffer'} = $buf;
	$EDITVIEW->{"buffer"}->RegisterAsObserver($EDITVIEW);
}

sub DispatchKeysToBuffer {
	#_DebugLog("View DispatchKeysToBuffer\n");
	do {
		my($sig) = shift @sigstack;
                if($sig eq "WINCH"){
                	View->ForceRedraw;
                }
		$key = &_GetKey(0);
                if($key != -1){
                        $EDITVIEW->{'buffer'}->DispatchKey($key);
                }
	} while($#sigstack >= 0 || $key != -1);
}

sub UpdateDisplay {
	#_DebugLog("View UpdateDisplay\n");
	$FULL->Display if(defined $FULL);
	$MINI->Display if(defined $MINI);

	if(!defined($MINI) or $MINI != $EDITVIEW){
	        Curses::move($HEIGHT-1,0);
		Curses::clrtoeol;
		Curses::addstr($EDITVIEW->{"buffer"}->{"curmsg"});
	}

        Curses::move($EDITVIEW->{"cursory"}+$EDITVIEW->{"line"},
		     $EDITVIEW->{"cursorx"}+$EDITVIEW->{"col"});
        Curses::refresh;
}

sub ForceRedraw {
	Curses::endwin;
        Curses::refresh;
        Curses::clear;
        Curses::getmaxyx($HEIGHT,$WIDTH);

	# Resize and position current views
	if(defined $FULL){
		$FULL->{"line"} = 0;
                $FULL->{"height"} = $HEIGHT-1;
                $FULL->{"col"} = 0;
                $FULL->{"width"} = $WIDTH;
	}
	if(defined $MINI){
                $MINI->{"line"} = $HEIGHT - 1;
                $MINI->{"height"} = 1;
                $MINI->{"col"} = 0;
                $MINI->{"width"} = $WIDTH;
        }    
}

sub ShowMsg { shift; _ShowMsg(@_); }
sub _ShowMsg {
        my($msg) = @_;
        Curses::move($HEIGHT-1,0);
        Curses::clrtoeol;
        Curses::addstr($msg);
        Curses::refresh;
}               

sub DisplayBanner { shift; &_DisplayBanner(@_); }
sub _DisplayBanner {
	my($line);

	Curses::clear;
	foreach $line (@_){
		Curses::move($y++,0);
		Curses::addstr($line);
	}
        Curses::refresh;
}

sub WaitForInput {
        $readset='';
        vec($readset,fileno(STDIN),1)=1;
        select($readset,undef,undef,undef);
}               

sub GetKey { shift; return &_GetKey(@_); }
sub _GetKey {
        my($blocking)=@_;
        &WaitForInput if($blocking);
        $key = Curses::getch;
        #getch sometimes returns a character and sometimes an integer
        #$key=ord($key) if($key != -1 && length($key)==1);
        $key=ord($key) if(length($key)==1);
        return $key;
}

# PUBLIC
sub PromptStr {
	shift;
        my($prompt,$str,@choices) = @_;
        #&_DebugLog("View PromptStr: $prompt $str");
        my($key);
        while(1){
                &_ShowMsg(_Expand($prompt . $str));
                $key=&_GetKey(1);
		if($key==$main::KeyCode{"^V"} ||
                   $key==$main::KeyCode{"^Q"}){
                	#Character Quote
			$key=&_GetKey(1);
			$str.=chr($key);
			next;
		}
                if($key==$main::KeyCode{"CR"})           { last; }
                if($key==$main::KeyCode{"DEL"})          { chop   $str; next;}
                if($key==$main::KeyCode{"KEY_BACKSPACE"}){ chop   $str; next;}
                $str.=chr($key);
        }
        #&DebugLog($str . "\n");
        return $str;
}

sub _Expand {
        my($str) = (@_);

        # put a carat before control chars
        $str =~ s/([\000-\037])/\^$1/g;
        $str =~ tr[\000-\037][\100-\137]; # make control chars viewable

        return($str);
}
sub catch_int {
        push @sigstack,@_;
}                                           
####################################
# Instance Methods

sub Notify {
	my($self) = shift;
	my($msg) = shift;

	#&main::DebugLog("Got notified $self\n");
	if($msg =~ /Deleted/i){
		my($buf) = $self->{'buffer'};
		#_DebugLog("View Notify: Unregistering $self from $buf\n");
		#$buf->UnregisterAsObserver($self);
		$MINI = undef if($self == $MINI);
		$FULL = undef if($self == $FULL);
		$EDITVIEW = undef if($self == $EDITVIEW);
	}
	if(!defined $EDITVIEW){
		$EDITVIEW = $MINI if(defined $MINI);
		$EDITVIEW = $FULL if(defined $FULL);
	}
}

sub BecomeEditView {
	my($self) = shift;
	#_DebugLog("View BecomeEditView $self\n");

	return if($self == $EDITVIEW);
	$self->{'buffer'}->MoveToXY($self->{'curcol'},$self->{'curline'});
	$EDITVIEW = $self;
}

sub Display {
	my($self) = shift;

	#Where the buffer says the cursor is
	my($dispcol,$displine) = $self->{'buffer'}->DispXY;

	#_DebugLog("View Display $dispcol $displine\n");

	# adjust vertically
        my($newtop) = $self->{'curtop'};
	my($height) = $self->{"height"}>1?$self->{"height"}-1:$self->{"height"};
	my($qheight) = int($height/4);
	$qheight = 1 if($qheight == 0);
        $newtop = int($displine/$qheight)*$qheight
		if($displine<$newtop);
        $newtop = (int(($displine - $height)/$qheight)+1)*$qheight
		if($displine>=$newtop+$height);
        $newtop = $displine if($newtop > $displine or
                               $newtop + $height <= $displine);
        $self->{'curtop'} = $newtop;                                    
	#_DebugLog("   newtop = $newtop\n");

	# adjust horizontally
        my($newleft) = $self->{'curleft'};
        my($width) = $self->{'width'};
        my($qwidth) = int($width/4);
	$qwidth = 1 if($qwidth == 0);
	$newleft = int($dispcol/$qwidth)*$qwidth
                if($dispcol<$newleft);
        $newleft = (int(($dispcol - $width)/$qwidth)+1)*$qwidth
                if($dispcol>=$newleft+$width);
        $newleft = $dispcol if($newleft > $dispcol or
                               $newleft + $width <= $dispcol);
        $self->{'curleft'} = $newleft;               
	#_DebugLog("   newleft = $newleft\n");

	my($buf) = $self->{'buffer'};
        my($statline,@display) =
		$buf->Display($newleft,$newtop,$self->{"width"},$height);

        for($line=0;$height > $line;$line++){
                Curses::move($self->{"line"}+$line,0);
                Curses::clrtoeol;
                Curses::addstr($display[$line]);
        }

	if($self->{"height"} >1){
	        Curses::move($self->{"line"}+$height,0);
		Curses::clrtoeol;
		Curses::addstr($statline);
	}
	if($self == $EDITVIEW){
		$self->{"cursorx"} = $dispcol-$newleft;
		$self->{"cursory"} = $displine-$newtop;
	}
}                                                       

sub _DebugLog { &main::DebugLog(@_); }

1;

